import json_line_logger
import subprocess
import tempfile
import os
import time
from .. import utils


def sbatch(
    script_path,
    script_arguments,
    stdout_path,
    stderr_path,
    jobname,
    logger=None,
    clusters=None,
    sbatch_path="sbatch",
    timeout=None,
):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    cmd = [sbatch_path]
    if clusters:
        cmd += ["--clusters", str.join(",", clusters)]

    cmd += ["--job-name", jobname]
    cmd += ["--output", stdout_path]
    cmd += ["--error", stderr_path]
    cmd += [script_path]
    for argument in script_arguments:
        cmd += [argument]

    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, timeout=timeout)
    except subprocess.CalledProcessError as e:
        logger.critical(str(e))
        raise


def scancel(
    jobname,
    scancel_path="scancel",
    timeout=None,
    timecooldown=1.0,
    logger=None,
):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    t = 0
    while True:
        try:
            _scancel(
                scancel_path=scancel_path,
                jobname=jobname,
                timeout=timeout,
                logger=logger,
            )
            break
        except KeyboardInterrupt:
            raise
        except Exception as bad:
            logger.warning("Problem in scancel()")
            logger.warning(str(bad))
            time.sleep(timecooldown)


def _scancel(scancel_path, jobname, timeout, logger=None):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    try:
        cmd = [scancel_path, "--name", str(jobname)]
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, timeout=timeout)
    except subprocess.CalledProcessError as e:
        logger.critical(str(e))
        raise


def squeue(
    squeue_path="squeue",
    timeout=None,
    timecooldown=1.0,
    logger=None,
    debug_dump_path=None,
):
    """
    Call slurm's squeue.

    Parameters
    ----------
    squeue_path : str, (default: "squeue")
        Path to the squeue executable
    timeout : float or None
        Time to wait for squeue to return.
    timecooldown : float
        Time to wait before calling squeue again in case of a problem.

    Returns
    -------
    squeue : list of dicts
        The jobs and their attributes
    """
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    numtry = 0
    while True:
        try:
            numtry += 1
            logger.debug("calling squeue, num. tries = {:d}".format(numtry))
            stdout = _squeue_format_all_stdout(
                squeue_path=squeue_path,
                timeout=timeout,
                logger=logger,
            )
            break
        except KeyboardInterrupt:
            logger.warning("KeyboardInterrupt")
            raise
        except Exception as bad:
            logger.warning("problem in _squeue_format_all_stdout()")
            logger.warning(str(bad))
            logger.warning("waiting for {:f}s".format(float(timecooldown)))
            time.sleep(timecooldown)

    logger.debug("parsing stdout into list of dicts")

    try:
        list_of_dicts = _parse_stdout_format_all(
            stdout=stdout,
            delimiter="|",
            logger=logger,
        )
        logger.debug("num. jobs in squeue = {:d}".format(len(list_of_dicts)))
    except Exception as err:
        logger.critical("Can not parse squeue's stdout.")
        if debug_dump_path:
            utils.write(path=debug_dump_path, content=stdout, mode="t")
            logger.critical("Dump stdout to {:s}.".format(debug_dump_path))
        raise err

    return list_of_dicts


def _parse_stdout_format_all(stdout, delimiter="|", logger=None):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    lines = str.splitlines(stdout)
    logger.debug("num. lines = {:d}".format(len(lines)))
    header_line = lines[0]

    keys = str.split(header_line, delimiter)
    keys = [str.lower(key) for key in keys]
    logger.debug("header line has {:d} keys".format(len(keys)))

    out = []
    # print("---lines---")
    for i in range(1, len(lines)):
        line = lines[i]
        # print("line: '{:s}'.".format(line))
        values = str.split(line, delimiter)
        if len(values) != len(keys):
            logger.debug("line {:d} has not expected num. of tokens".format(i))
        line_dict = {}
        for j in range(len(keys)):
            jkey = keys[j]
            jval = values[j]
            line_dict[jkey] = jval
        out.append(line_dict)
    return out


def _squeue_format_all_stdout(squeue_path="squeue", timeout=None, logger=None):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    with tempfile.TemporaryDirectory(prefix="slurmpypoolurm") as tmp:
        tmp_stdout_path = os.path.join(tmp, "stdout.txt")

        logger.debug("stdout in {:s}".format(tmp_stdout_path))
        if timeout:
            logger.debug("timeout = {:f}s".format(float(timeout)))

        with open(tmp_stdout_path, "wt") as f:
            p = subprocess.Popen(
                [squeue_path, "--format", "%all"],
                stdout=f,
            )
            p.wait(timeout=timeout)

        with open(tmp_stdout_path, "rt") as f:
            stdout = f.read()

    logger.debug("len(stdout) = {:d}".format(len(stdout)))
    return stdout



#!/usr/bin/env python3
import argparse
import json
import datetime
import sys
import pypoolparty

qpaths = pypoolparty.slurm.testing.dummy_paths()

# dummy sbatch
# ============
parser = argparse.ArgumentParser(description="dummy slurm sbatch")
parser.add_argument("--clusters", type=str, help="CSV of clusters")
parser.add_argument("--output", type=str, help="stdout path")
parser.add_argument("--error", type=str, help="stderr path")
parser.add_argument("--job-name", type=str, help="jobname")
parser.add_argument("script_args", nargs="*", default=None)

args = parser.parse_args()

assert len(args.script_args) == 2

with open(qpaths["queue_state"], "rt") as f:
    state = json.loads(f.read())

now = datetime.datetime.now()
jobid = str(int(now.timestamp() * 1e6))

_worker_node_script_path = args.script_args[0]
_python_path = pypoolparty.testing.read_shebang_path(
    path=_worker_node_script_path
)

job = {
    "STATE": "PENDING",
    "JOBID": jobid,
    "NAME": args.job_name,
    "REASON": "foobar",
    "_opath": args.output,
    "_epath": args.error,
    "_python_path": _python_path,
    "_script_arg_0": args.script_args[0],
    "_script_arg_1": args.script_args[1],
}

state["pending"].append(job)

with open(qpaths["queue_state"], "wt") as f:
    f.write(json.dumps(state, indent=4))

sys.exit(0)



from . import utils
import glob
import os


def list_ids(work_dir, wildcard, num_digits_jobid=9):
    paths = glob.glob(os.path.join(work_dir, wildcard))
    basenames = [os.path.basename(path) for path in paths]
    list_ids = [int(basename[0:num_digits_jobid]) for basename in basenames]
    return list_ids


def make_summary(work_dir, max_osize=1000 * 1000, max_esize=1000 * 1000):
    col = {}
    print("make list started")
    col["started"] = set(list_ids(work_dir, "*.pkl"))
    print("make list completed")
    col["completed"] = set(list_ids(work_dir, "*.pkl.out"))
    col["incomplete"] = col["started"].difference(col["completed"])
    col["stdout"] = {}
    col["stderr"] = {}
    print("read stdout and stderr")

    for c, jid in enumerate(col["started"]):
        if c % (79 - 13) == 0:
            print("{:6d}/{:6d}".format(c + 1, len(col["started"])))
        print(".", end="", flush=True)
        try:
            opath = os.path.join(work_dir, "{:09d}.pkl.o".format(jid))
            epath = os.path.join(work_dir, "{:09d}.pkl.e".format(jid))
            osize = os.stat(opath).st_size
            osize_read = min([osize, max_osize])
            with open(opath, "rb") as f:
                col["stdout"][jid] = f.read(osize_read)
            esize = os.stat(epath).st_size
            esize_read = min([esize, max_esize])
            with open(epath, "rb") as f:
                col["stderr"][jid] = f.read(esize_read)
        except Exception as error:
            print("Failed to read stdout and stderr")
            print(error)

    print("read jobs of 'incomplete'")
    col["incomplete_jobs"] = {}

    for jid in col["incomplete"]:
        try:
            jpath = os.path.join(work_dir, "{:09d}.pkl".format(jid))
            job = utils.read_pickle(path=jpath)
            col["incomplete_jobs"][jid] = job
        except Exception as error:
            print("Failed to read incomplete job's *.pkl.")
            print(error)

    return col



import subprocess
import qstat as external_qstat_call
import time


def qsub(
    qsub_path,
    queue_name,
    script_exe_path,
    script_path,
    arguments,
    JB_name,
    stdout_path,
    stderr_path,
    logger,
):
    cmd = [qsub_path]
    if queue_name:
        cmd += ["-q", queue_name]
    cmd += ["-o", stdout_path]
    cmd += ["-e", stderr_path]
    cmd += ["-N", JB_name]
    cmd += ["-S", script_exe_path]
    cmd += [script_path]
    for argument in arguments:
        cmd += [argument]

    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        logger.critical("Error in qsub()")
        logger.critical("qsub() returncode: {:d}".format(e.returncode))
        logger.critical(e.output)
        raise


def _qdel(JB_job_number, qdel_path, logger):
    try:
        _ = subprocess.check_output(
            [qdel_path, str(JB_job_number)],
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as e:
        logger.critical("qdel returncode: {:s}".format(e.returncode))
        logger.critical("qdel stdout: {:s}".format(e.output))
        raise


def qdel(JB_job_number, qdel_path, logger):
    while True:
        try:
            _qdel(JB_job_number, qdel_path, logger=logger)
            break
        except KeyboardInterrupt:
            raise
        except Exception as bad:
            logger.warning("Problem in qdel()")
            logger.warning(str(bad))
            time.sleep(1)


def qstat(qstat_path, logger):
    """
    Return lists of running and pending jobs.
    Try again in case of Failure.
    Only except KeyboardInterrupt to stop.
    """
    while True:
        try:
            running, pending = external_qstat_call.qstat(qstat_path=qstat_path)
            break
        except KeyboardInterrupt:
            raise
        except Exception as bad:
            logger.warning("Problem in qstat()")
            logger.warning(str(bad))
            time.sleep(1)
    return running, pending



import json
import subprocess
from . import utils


def dummy_init_queue_state(path, evil_jobs=[]):
    qstate = {"running": [], "pending": [], "evil_jobs": evil_jobs}
    utils.write_text(path=path, content=json.dumps(qstate))


def dummy_run_job(job):
    with open(job["_opath"], "wt") as o, open(job["_epath"], "wt") as e:
        subprocess.call(
            [job["_python_path"], job["_script_arg_0"], job["_script_arg_1"]],
            stdout=o,
            stderr=e,
        )


def read_shebang_path(path):
    txt = utils.read_text(path=path)
    lines = str.splitlines(txt)
    firstline = lines[0]
    assert str.startswith(firstline, "#!")
    return firstline[2:]



#!/usr/bin/env python3
import argparse
import json
import datetime
import sys
from pypoolparty import sun_grid_engine

qpaths = sun_grid_engine.testing.dummy_paths()


# dummy qsub
# ==========
parser = argparse.ArgumentParser(description="dummy sun-grid-engine qsub")
parser.add_argument("-q", type=str, help="Name of queue")
parser.add_argument("-o", type=str, help="stdout path")
parser.add_argument("-e", type=str, help="stderr path")
parser.add_argument("-N", type=str, help="JB_name")
parser.add_argument("-V", action="store_true", help="export environment")
parser.add_argument("-S", type=str, help="path of script")
parser.add_argument("script_args", nargs="*", default=None)

args = parser.parse_args()

assert len(args.script_args) == 2

with open(qpaths["queue_state"], "rt") as f:
    state = json.loads(f.read())

now = datetime.datetime.now()
JB_job_number = str(int(now.timestamp() * 1e6))

job = {
    "@state": "pending",
    "JB_job_number": JB_job_number,
    "JAT_prio": "0.50500",
    "JB_name": args.N,
    "JB_owner": "dummy_user",
    "state": "qw",
    "JB_submission_time": now.isoformat(),
    "queue_name": str(args.q),
    "slots": "1",
    "_opath": args.o,
    "_epath": args.e,
    "_python_path": args.S,
    "_script_arg_0": args.script_args[0],
    "_script_arg_1": args.script_args[1],
}

state["pending"].append(job)

with open(qpaths["queue_state"], "wt") as f:
    f.write(json.dumps(state, indent=4))

sys.exit(0)



#!/usr/bin/env python3
import sys
import json
import datetime
import subprocess
import pypoolparty

qpaths = pypoolparty.sun_grid_engine.testing.dummy_paths()


def job_to_xml(job):
    jld = ""
    jld += '<job_list state="{:s}">\n'.format(job["@state"])
    jld += "    <JB_job_number>{:s}</JB_job_number>\n".format(
        job["JB_job_number"]
    )
    jld += "    <JAT_prio>{:s}</JAT_prio>\n".format(job["JAT_prio"])
    jld += "    <JB_name>{:s}</JB_name>\n".format(job["JB_name"])
    jld += "    <JB_owner>{:s}</JB_owner>\n".format(job["JB_owner"])
    jld += "    <state>{:s}</state>\n".format(job["state"])
    jld += "    <JB_submission_time>{:s}</JB_submission_time>\n".format(
        job["JB_submission_time"]
    )
    jld += "    <queue_name>{:s}</queue_name>\n".format(job["queue_name"])
    jld += "    <slots>{:s}</slots>\n".format(job["slots"])
    jld += "</job_list>\n"
    return jld


def state_to_xml(state):
    out_xml = "<?xml version='1.0'?>\n"
    out_xml += "<job_info>\n"

    out_xml += "    <queue_info>\n"
    for job in state["running"]:
        out_xml += indent_text(job_to_xml(job), indent=8)
    out_xml += "    </queue_info>\n"

    out_xml += "    <job_info>\n"
    for job in state["pending"]:
        out_xml += indent_text(job_to_xml(job), indent=8)
    out_xml += "    </job_info>\n"

    out_xml += "</job_info>\n"
    return out_xml


def indent_text(text, indent=4):
    out = []
    spaces = " " * indent
    for line in text.splitlines():
        out.append(spaces + line + "\n")
    return "".join(out)


# dummy qstat
# ===========
# Every time this is called, it runs one job.
MAX_NUM_RUNNING = 10

assert len(sys.argv) == 2
assert sys.argv[1] == "-xml"

with open(qpaths["queue_state"], "rt") as f:
    state = json.loads(f.read())

evil_ichunks_num_fails = {}
evil_ichunks_max_num_fails = {}
for evil in state["evil_jobs"]:
    evil_ichunks_num_fails[evil["ichunk"]] = evil["num_fails"]
    evil_ichunks_max_num_fails[evil["ichunk"]] = evil["max_num_fails"]


if len(state["running"]) >= MAX_NUM_RUNNING:
    run_job = state["running"].pop(0)
    pypoolparty.testing.dummy_run_job(run_job)
elif len(state["pending"]) > 0:
    job = state["pending"].pop(0)
    ichunk = pypoolparty.pooling.make_ichunk_from_jobname(
        jobname=job["JB_name"]
    )
    if ichunk in evil_ichunks_num_fails:
        if evil_ichunks_num_fails[ichunk] < evil_ichunks_max_num_fails[ichunk]:
            job["@state"] = "?"
            job["state"] = "Eqw"
            state["pending"].append(job)
            evil_ichunks_num_fails[ichunk] += 1
        else:
            job["@state"] = "running"
            job["state"] = "r"
            state["running"].append(job)
    else:
        job["@state"] = "running"
        job["state"] = "r"
        state["running"].append(job)
elif len(state["running"]) > 0:
    run_job = state["running"].pop(0)
    pypoolparty.testing.dummy_run_job(run_job)


evil_jobs = []
for ichunk in evil_ichunks_num_fails:
    evil_jobs.append(
        {
            "ichunk": ichunk,
            "num_fails": evil_ichunks_num_fails[ichunk],
            "max_num_fails": evil_ichunks_max_num_fails[ichunk],
        }
    )
state["evil_jobs"] = evil_jobs


with open(qpaths["queue_state"], "wt") as f:
    f.write(json.dumps(state, indent=4))

out_xml = state_to_xml(state)
print(out_xml)

sys.exit(0)



#!/usr/bin/env python3
import sys
import json
import datetime
from pypoolparty import sun_grid_engine

qpaths = sun_grid_engine.testing.dummy_paths()

# dummy qdel
# ==========
assert len(sys.argv) == 2
JB_job_number = sys.argv[1]

with open(qpaths["queue_state"], "rt") as f:
    old_state = json.loads(f.read())

found = False
state = {
    "pending": [],
    "running": [],
    "evil_jobs": old_state["evil_jobs"],
}
for job in old_state["running"]:
    if job["JB_job_number"] == JB_job_number:
        found = True
    else:
        state["running"].append(job)

for job in old_state["pending"]:
    if job["JB_job_number"] == JB_job_number:
        found = True
    else:
        state["pending"].append(job)

with open(qpaths["queue_state"], "wt") as f:
    f.write(json.dumps(state, indent=4))

if found == True:
    sys.exit(0)
else:
    print("Can not find ", JB_job_number)
    sys.exit(1)
